#ifndef SEQ
#include "mpi.h"
#endif

#include "moldyn.h"

int input_parameters (rank, xid, yid, zid, xor, yor, zor)
int    rank;
int   *xid;
int   *yid;
int   *zid;
float *xor;
float *yor;
float *zor;
{
   extern float sigma, rcut, dt, eqtemp, sfx, sfy, sfz, cellix, celliy, celliz;
   extern float freex, boxlx, boxly, boxlz, sr6, vrcut, dvrcut, dvrc12;
   extern int nprosx, nprosy, nprosz, nstep, nequil, nstart;
   extern int iscale, isave, isvunf, isvpfs, nc, natoms, mx, my, mz;

   float pass[20];
   int abort;

   if (I_AM_HOST) abort = read_input (pass);

#ifndef SEQ
   MPI_Bcast (&abort, 1, MPI_INT,  RANKHOST, MPI_COMM_WORLD);
#endif

   if (abort==0){
#ifndef SEQ
      MPI_Bcast (pass, 16, MPI_REAL, RANKHOST, MPI_COMM_WORLD);
#endif

      sigma  = pass[0];
      rcut   = pass[1];
      dt     = pass[2];
      nprosx = (int)(pass[3]+0.1);
      nprosy = (int)(pass[4]+0.1);
      nprosz = (int)(pass[5]+0.1);
      nstep  = (int)(pass[6]+0.1);
      nequil = (int)(pass[7]+0.1);

      if (pass[8] < 0.0)
         isave = (int)(pass[8]-0.1);
      else
         isave = (int)(pass[8]+0.1);

      iscale = (int)(pass[9]+0.1);
      eqtemp = pass[10];
      nstart = (int)(pass[11]+0.1);

      if (pass[12] < 0.0)
         isvunf = (int)(pass[12]-0.1);
      else
         isvunf = (int)(pass[12]+0.1);

      if (pass[13] < 0.0)
         isvpfs = (int)(pass[13]-0.1);
      else
         isvpfs = (int)(pass[13]+0.1);

      nc    = (int)(pass[14]+0.1);
      freex = pass[15];

      natoms = 4*nc*nc*nc;

      *xid = rank%nprosx;
      *yid = (rank/nprosx)%nprosy;
      *zid = rank/(nprosx*nprosy);

      *xor = ((float)(*xid)+0.5)/(float)nprosx - 0.5;
      *yor = ((float)(*yid)+0.5)/(float)nprosy - 0.5;
      *zor = ((float)(*zid)+0.5)/(float)nprosz - 0.5;

      boxlx = 1.0/(float)nprosx;
      boxly = 1.0/(float)nprosy;
      boxlz = 1.0/(float)nprosz;

      mx = min(MMX, (int)(boxlx/rcut)+2);
      my = min(MMX, (int)(boxly/rcut)+2);
      mz = min(MMX, (int)(boxlz/rcut)+2);

      sfx = ((float)mx - 2.0)/((float)mx*boxlx);
      sfy = ((float)my - 2.0)/((float)my*boxly);
      sfz = ((float)mz - 2.0)/((float)mz*boxlz);

      cellix = (float)mx;
      celliy = (float)my;
      celliz = (float)mz;

      set_nearest_cells ();

      sr6    = (sigma/rcut)*(sigma/rcut);
      sr6    = sr6*sr6*sr6;
      vrcut  = sr6*(sr6-1.0);
      dvrcut = -1.0*sr6*(sr6-0.5)/rcut;
      dvrc12 = 12.0*dvrcut;
   }
   return (abort);
}
